//
// Created by landon on 8/8/22.
//
#pragma once
#include "model_utils.hpp"

#if false
namespace PYSIMLINK{
    void print_model_params(const rtwCAPI_ModelMappingInfo *mmi);
    void print_block_params(const rtwCAPI_ModelMappingInfo *mmi);
    void print_signals(const rtwCAPI_ModelMappingInfo *mmi);
    void print_params_recursive(const rtwCAPI_ModelMappingInfo *child_mmi);
};

}
#endif